import typing
import inspect

from inspect import Parameter
from datetime import timedelta
from collections import namedtuple

from toolkit.singleton import Singleton
from toolkit.frozen import FrozenSettings

from werkzeug.http import parse_options_header
from werkzeug.datastructures import ImmutableMultiDict
from flask.sessions import SecureCookieSessionInterface

from apistar.server.asgi import ASGIReceive
from apistar.conneg import negotiate_content_type
from apistar.server.validation import RequestDataComponent
from apistar import Route, exceptions, http, Component as _Component

from apistellar.types import Type

from .controller import Controller
from .entities import Session, Cookie, FormParam, FileStream, DummyFlaskApp, \
    SettingsMixin, MultiPartForm, UrlEncodeForm


class Component(_Component):

    def identity(self, parameter: inspect.Parameter):
        """
        修复annotation_name重名的Bug
        """
        parameter_name = parameter.name.lower()
        annotation_name = str(parameter.annotation)

        # If `resolve_parameter` includes `Parameter` then we use an identifier
        # that is additionally parameterized by the parameter name.
        args = inspect.signature(self.resolve).parameters.values()
        if inspect.Parameter in [arg.annotation for arg in args]:
            return annotation_name + ':' + parameter_name

        # Standard case is to use the class name, lowercased.
        return annotation_name

    def resolve(self, *args, **kwargs) -> object:
        raise NotImplementedError()

    def can_handle_parameter(self, parameter: inspect.Parameter):
        """重写这个方法是为了增加typing.Union类型的判定"""
        return_annotation = inspect.signature(self.resolve).return_annotation
        if return_annotation is inspect.Signature.empty:
            msg = (
                'Component "%s" must include a return annotation on the '
                '`resolve()` method, or override `can_handle_parameter`'
            )
            raise exceptions.ConfigurationError(msg % self.__class__.__name__)
        return type(return_annotation) == typing._Union and \
               parameter.annotation in return_annotation.__args__ or \
               parameter.annotation is return_annotation


class ServiceComponent(Component, metaclass=Singleton):
    """
    注入Service
    """
    def resolve(self, route: Route) -> Controller:
        return route.controller


class SettingsComponent(Component, SettingsMixin):
    """
    注入Settings
    """
    def resolve(self) -> FrozenSettings:
        return self.settings


class CookiesComponent(Component):
    def resolve(self, cookie: http.Header) -> typing.Dict[str, Cookie]:
        cookies = dict()
        if cookie:
            for c in cookie.split(";"):
                key, val = c.strip().split("=", 1)
                cookies[key] = Cookie(val)
        return cookies


class CookieComponent(Component):

    def resolve(self,
                parameter: Parameter,
                cookies: typing.Dict[str, Cookie]) -> Cookie:
        return cookies.get(parameter.name.replace('_', '-'))


class DummyFlaskAppComponent(Component):

    def __init__(self):
        self.default_config = {
            'SESSION_COOKIE_NAME': 'session',
            'SESSION_COOKIE_DOMAIN': None,
            'SESSION_COOKIE_PATH': '/',
            'SESSION_COOKIE_HTTPONLY': True,
            'SESSION_COOKIE_SECURE': False,
            'SESSION_COOKIE_SAMESITE': None,
            'SESSION_REFRESH_EACH_REQUEST': True,
            'SERVER_NAME': None,
        }

    def resolve(self, settings: FrozenSettings,
        host: http.Header,
        ip_host: http.Host) -> DummyFlaskApp:
        config = self.default_config.copy()
        config["SERVER_NAME"] = host or ip_host
        return DummyFlaskApp(
            config=settings.get("SESSION_CONFIG", config),
            secret_key=settings.get("SECRET_KEY"),
            permanent_session_lifetime=timedelta(
                days=settings.get("PERMANENT_SESSION_LIFETIME", 31)),
            session_cookie_name=settings.get(
                "SESSION_COOKIE_NAME", "session"))


class SessionComponent(Component):

    def __init__(self):
        self.dummy_request = namedtuple("Request", "cookies")
        self.session_interface = SecureCookieSessionInterface()

    def resolve(self,
                app: DummyFlaskApp,
                cookies: typing.Dict[str, Cookie]) -> Session:
            request = self.dummy_request(cookies=cookies)
            return self.session_interface.open_session(app, request)


class FileStreamComponent(Component):
    media_type = 'multipart/form-data'

    async def decode(self, receive, headers):
        try:
            mime_type, mime_options = parse_options_header(
                headers['content-type'])
        except KeyError:
            mime_type, mime_options = '', {}

        boundary = mime_options.get('boundary', "").encode()
        if boundary is None:
            raise ValueError('Missing boundary')

        return FileStream(receive, boundary)

    async def resolve(self,
                      receive: ASGIReceive,
                      headers: http.Headers,
                      content_type: http.Header) -> FileStream:
        try:
            negotiate_content_type([self], content_type)
        except exceptions.NoCodecAvailable:
            raise exceptions.UnsupportedMediaType()

        return await self.decode(receive, headers)


class FormParamComponent(Component):

    def resolve(self, parameter: inspect.Parameter,
                form: http.RequestData) -> FormParam:
        if parameter.name in (form or {}):
            return FormParam(form[parameter.name])


class ValidateRequestDataComponent(_Component):
    """
    当不存在可以处理model的factory时，使用这个类来接管，将request data生成model对象
    """
    def identity(self, parameter: inspect.Parameter):
        """
        修复annotation_name重名的Bug
        """
        parameter_name = parameter.name.lower()
        annotation_name = str(parameter.annotation)
        # If `resolve_parameter` includes `Parameter` then we use an identifier
        # that is additionally parameterized by the parameter name.
        args = inspect.signature(self.resolve).parameters.values()
        if inspect.Parameter in [arg.annotation for arg in args]:
            return annotation_name + ':' + parameter_name

        # Standard case is to use the class name, lowercased.
        return annotation_name

    def can_handle_parameter(self, parameter: inspect.Parameter):
        return issubclass(parameter.annotation, Type)

    def resolve(self,
                route: Route,
                data: http.RequestData):
        body_field = route.link.get_body_field()
        if not body_field or not body_field.schema:
            return data

        if isinstance(data, ImmutableMultiDict):
            data = self._change_to_dict(data)
        validator = body_field.schema
        return validator.model(data)

    @staticmethod
    def _change_to_dict(data):
        """
        由于form表单使用`werkzeug.datastructures.ImmutableMultiDict`来保存form中的
        key-val, 当key只对应一个val时，我们希望得到key-val1这种形式，如果key对应多个val
        我们希望得到key-[val1, val2...]这种形式。
        :param data:
        :return:
        """
        return {k: v if len(v) > 1 else v[0] or None for k, v in dict(data).items()}


class MultiPartComponent(RequestDataComponent, Component):

    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation is MultiPartForm


class UrlEncodeComponent(RequestDataComponent, Component):

    def can_handle_parameter(self, parameter: inspect.Parameter):
        return parameter.annotation is UrlEncodeForm



"""
PYTEST_DONT_REWRITE
"""
from apistellar.app import Application, show_routes

from apistellar.build import main
from apistellar.build.tasks import Task

from apistellar.bases.service import Service
from apistellar.bases.hooks import Hook, Return
from apistellar.bases.components import Component
from apistellar.bases.controller import Controller
from apistellar.bases.response import FileResponse
from apistellar.bases.model_factory import ModelFactory
from apistellar.bases.entities import Session, Cookie, FormParam, \
    FileStream, inject, SettingsMixin, UrlEncodeForm, MultiPartForm

from apistellar.console import main as console

from apistellar.solo import Solo
from apistellar.solo.manager import SoloManager

from apistellar.persistence import DriverMixin, conn_manager, \
    conn_ignore, proxy, contextmanager

from apistellar.types import Type, AsyncType, PersistentType, \
    TypeEncoder, validators

from apistellar.route import route, get, post, delete, put, options, websocket

from apistellar.helper import redirect, require, return_wrapped

__version__ = "1.1.5"     



import re
import os
import sys
import string

from collections import OrderedDict
from importlib import import_module
from os import makedirs, sep, listdir, getcwd
from os.path import join, exists, abspath, dirname, isdir, basename

from toolkit import load
from toolkit.markdown_helper import MarkDownRender
from apistellar.types import validators
from apistellar.document import DocPainter

__all__ = ["Task", "Project", "Service", "Model", "Solo", "Document"]


class Task(object):
    def __init__(self):
        self.template = None
        self.kwargs = {}

    def create(self, env, **kwargs):
        task = kwargs.pop("task")
        names = kwargs.pop("name", [])
        self.kwargs.update(kwargs)

        for name in names:
            self.enrich_kwargs(self.validate_name(name))
            makedirs(self.kwargs["dirname"], exist_ok=True)
            self.copytree(env, task)
        print("、".join(names), "已创建。")

    def validate_name(self, name):
        return name

    def enrich_kwargs(self, name):
        self.kwargs["name"] = name
        self.kwargs["dirname"] = name

    @classmethod
    def enrich_parser(cls, sub_parser):
        pass

    def copytree(self, env, task, dest_path=None):
        if dest_path is None:
            dest_path = self.kwargs["dirname"]
        copy_path = join(self.template, task)

        for file in listdir(copy_path):
            file = join(copy_path, file)

            if file.count("__pycache__"):
                continue

            if isdir(file):
                dir_name = abspath(join(dest_path, self.render_path_name(file)))
                makedirs(dir_name, exist_ok=True)
                self.copytree(env, file, dir_name)
            else:
                template = env.get_template(file.replace(self.template, ""))
                file = self.render_path_name(file)
                filename = abspath(join(dest_path, file).replace(".tmpl", ""))
                if exists(filename) and \
                        input(f"{filename}已存在，是否覆盖y/n?") not in ["y", "yes"]:
                    continue

                with open(filename, "w") as f:
                    f.write(template.render(**self.kwargs))
                    f.write("\n")

    def render_path_name(self, path):
        return string.Template(basename(path)).substitute(**self.kwargs)


class Project(Task):
    """
    项目
    """
    def validate_name(self, name):
        assert re.search(r'^[_a-zA-Z]\w*$', name), \
            '项目名称只能包含数字、字母、下划线并以字母开头！'
        try:
            mod = import_module(name)
        except ImportError:
            mod = None
        assert mod is None, f'模块 {name} 已存在！'
        return name

    def enrich_kwargs(self, name):
        super(Project, self).enrich_kwargs(name)
        self.kwargs["project_dir"] = join(getcwd(), name)

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs=1, help="项目名称")


class ModuleTask(Task):

    def validate_name(self, name):
        words = re.findall(r"([A-Za-z0-9]+)", name)
        assert words, f"name: {name} invalid!"
        assert words[0][0].isalpha(), f"name: {name} start with number!"
        return words

    def enrich_kwargs(self, words):
        class_name = "".join(word.capitalize() for word in words)
        module_name = "_".join(word.lower() for word in words)
        self.kwargs["class_name"] = class_name
        self.kwargs["module_name"] = module_name
        self.kwargs["dirname"] = module_name


class Service(ModuleTask):
    """
    服务
    """

    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        father = None

        if exists("__init__.py"):
            regex = re.compile(r"class\s+(\w*?Controller)\(\w*Controller\):")
            mth = regex.search(open("__init__.py").read())
            if mth:
                father = mth.group(1)
        self.kwargs["father"] = father or "Controller"

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs="+", help="服务模块名称")


class Drivable(Task):

    def enrich_kwargs(self, name):
        super().enrich_kwargs(name)
        driver = self.kwargs.get("driver")
        if driver and driver.count(":") == 1:
            module, cls = driver.split(":")
        else:
            module, cls = None, None
        self.kwargs["module"] = module
        self.kwargs["cls"] = cls

    @classmethod
    def enrich_parser(cls, sub_parser):
        super().enrich_parser(sub_parser)
        sub_parser.add_argument(
            "-d", "--driver", help="驱动 eg：pymongo:MongoClient")


class Model(ModuleTask):
    """
    模型层
    """
    @staticmethod
    def to_list(val):
        """用来兼容接口"""
        return [val]

    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        fields = [f.split(":", 1) for f in self.kwargs.get("fields", [])]
        types = dict()

        for p in dir(validators):
            obj = getattr(validators, p)
            if isinstance(obj, type) and \
                    issubclass(obj, validators.Validator):
                types[obj.__name__.lower()] = obj.__name__

        new_fields = []
        for k, v in fields[:]:
            new_fields.append((k, types.get(v.lower(), "")))
        self.kwargs["fields"] = new_fields
        self.kwargs["dirname"] = self.kwargs.get("path") or self.kwargs["dirname"]

    @classmethod
    def enrich_parser(cls, sub_parser):
        super().enrich_parser(sub_parser)
        sub_parser.add_argument(
            "-n", "--name", required=True, type=cls.to_list, help="models名称")
        sub_parser.add_argument(
            "-p", "--path", help="所属服务路径 eg: article/comment")
        sub_parser.add_argument(
            "-a", "--async", action="store_true", help="是否拥有异步获取属性的能力")
        sub_parser.add_argument("fields", nargs="*", help="字段 eg: id:integer")


class Solo(ModuleTask):
    """
    独立任务程序
    """
    def enrich_kwargs(self, words):
        super().enrich_kwargs(words)
        project_name = basename(abspath(getcwd()))
        self.kwargs["project_name"] = project_name
        self.kwargs["dirname"] = join(project_name, self.kwargs["dirname"])
        self.kwargs["back_trace"] = ".." + sep + ".." + sep

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs="+", help="独立任务程序名称")


class Document(Task):

    def create(self, env, **kwargs):
        task = kwargs.pop("task")
        parser = kwargs.pop("parser")
        name = kwargs.pop("name", [])[0]
        module = kwargs["module"]
        location = abspath(join(os.getcwd(), kwargs.pop("location")))


        if module:
            current_dir = dirname(load(module).__file__)
            sys.modules.pop(module, None)
        else:
            current_dir = "."
        os.chdir(current_dir)

        painter = DocPainter(current_dir, parser)
        # 不再提醒是否覆盖
        global input
        input = self._input
        base_dir_name = join(location, name)

        indices = OrderedDict()
        for parent, doc in painter.paint().items():
            doc["dirname"] = base_dir_name
            self.kwargs = doc
            fn = os.path.join(
                base_dir_name, doc["file_path"], doc["doc_name"] + ".md.html")
            indices[fn] = doc["doc_name"]
            self.copytree(env, task)

        os.makedirs(base_dir_name, exist_ok=True)
        with open(os.path.join(base_dir_name, "index.md"), "w") as f:
            f.write(f"# {name}文档\n\n")
            for index, (key, val) in enumerate(indices.items()):
                f.write(f"{index+1}. [{val}]({key})\n")

        print(f"{name}已创建。")
        with MarkDownRender("github-markdown.css", base_dir_name) as mk_render:
            output = mk_render.render()

        if output:
            os.system(f"open {output}")
        else:
            print("未找到可打开的文档！")

    @staticmethod
    def _input(prompt):
        return "y"

    @classmethod
    def enrich_parser(cls, sub_parser):
        sub_parser.add_argument("name", nargs=1, help="文档名称")
        sub_parser.add_argument("-m", "--module", help="模块地址", required=True)
        sub_parser.add_argument("-l", "--location", help="文章地址", default=".")
        sub_parser.add_argument(
            "-p", "--parser", help="parser模块地址",
            default="apistellar.document.parser.RstDocParserDocParser")




import re
import os
import json
import typing
import inspect

from functools import reduce
from collections import OrderedDict
from abc import ABC, abstractmethod

from toolkit import cache_method
from apistar import Include, http

from apistellar.types import Type
from apistellar.bases.entities import FormParam, FileStream, \
    MultiPartForm, UrlEncodeForm


class DummyParam(object):
    def __init__(self, name, default, annotation):
        self.name = name
        if default is inspect._empty:
            self.default = default
        else:
            self.default = default() if callable(default) else default
        self.annotation = annotation


class Parser(ABC):

    @abstractmethod
    def parse_docs(self, routes):
        """
        通过route带来的信息，解析成接口文档信息
        :param routes:
        :return:
        """

    def walk_route(self, routes, parents=None):
        """
        遍历所有的include，找到route及其patient include
        :param routes:
        :param parents:
        :return:
        """
        for route in routes:
            if parents is None:
                current_parents = []
            else:
                current_parents = parents[:]
            if isinstance(route, Include):
                current_parents.append(route)
                yield from self.walk_route(route.routes, current_parents)
            else:
                yield route, parents

    @staticmethod
    def _extract_pattern(route, parents):
        return reduce(
            lambda x, y: x.rstrip('/') + y.url, parents, "").rstrip("/") \
               + route.url


class LogParser(Parser):

    def parse_docs(self, routes):
        for route, parents in self.walk_route(routes):
            name = reduce(lambda x, y: f"{x}:{y.name}", parents[1:], "view") \
                   + ":" + route.name
            cont = route.controller.__class__
            ca_name = f"{cont.__module__}:{cont.__name__}" \
                      f"#{route.handler.__name__}"
            pattern = self._extract_pattern(route, parents)
            yield route.method, pattern, name, ca_name


class RstDocParserDocParser(Parser):
    """
    用于解析rst风格的注释，并生成文档json，用于渲染文档
    """
    param_annotation_mapping = {
        http.QueryParam: "str",
        str: "str",
        int: "int",
        bool: "bool",
        float: "float",
        http.PathParam: "str",
        FormParam: "str",
        http.RequestData: "json",
        FileStream: "file",
        MultiPartForm: "file",
        UrlEncodeForm: "dict",
        http.QueryParams: "dict"
    }
    # 支持哪几种注释类型
    types = ("ex", "param", "type", "return")
    regex_tail = f"(?=(?:{'|'.join('(?::%s)' % type for type in types)}|$))"

    def parse_docs(self, routes):
        docs = OrderedDict()

        for route, parents in self.walk_route(routes):
            ps = tuple(x.name for x in parents[1:])
            doc = docs.setdefault(ps, dict())
            doc.setdefault("parents", ps)
            doc_name = route.controller.__class__.__doc__
            doc_name = doc_name.strip() if doc_name else \
                route.controller.__class__.__name__
            doc.setdefault("doc_name", doc_name.strip())
            all_type_info = doc.setdefault("type_info", dict())
            interfaces = doc.setdefault("interfaces", list())
            info = self._extract_info(route, parents)
            all_type_info.update(info.pop("type_info", dict()))
            interfaces.append(info)

            doc.setdefault("enumerate", enumerate)
            doc.setdefault("bool", bool)
            doc.setdefault("iter", self.iter_interface)
            doc.setdefault("len", len)
            doc.setdefault("map", map)
            doc.setdefault("file_path", os.path.join(*ps))
        return docs

    @staticmethod
    def iter_interface(interface):
        if "params" in interface:
            yield ("查询参数", interface["params"])

        if "path_params" in interface:
            yield ("路径参数", interface["path_params"])

        if "form_params" in interface:
            yield ("表单参数", interface["form_params"])

    @classmethod
    def regex_format(cls, type, param_name, sub_reg="(.*?)"):
        return r":{} +{}:{}\s+?{}".format(
            type, param_name, sub_reg, cls.regex_tail)

    def _extract_info(self, route, parents):
        info = dict()
        handler = route.handler
        type_info = dict()
        info["endpoint"] = self._extract_pattern(route, parents)
        info["method"] = route.method
        info["comment"] = self._extract_comment(handler.__doc__)
        params, form_params, path_params = (OrderedDict() for i in range(3))
        json_params = None
        sign = inspect.signature(handler)
        path_names = {
            item.strip('{}').lstrip('+'): item.strip('{}')
            for item in re.findall('{[^}]*}', route.url)
        }

        for name, param in sign.parameters.items():
            if param.annotation in self.param_annotation_mapping.keys():
                attributes = self._get_params_attr(param, handler.__doc__)

                if name in path_names.keys():
                    if path_names[name].startswith("+"):
                        attributes["type"] = "path"
                    # path_params不能有默认参数
                    if "default" in attributes:
                        del attributes["default"]
                    path_params[name] = attributes
                elif param.annotation == FormParam:
                    form_params[name] = attributes
                elif attributes["type"] == "json":
                    json_params = attributes
                elif attributes["type"] == "file":
                    form_params["`file1`, `file2`, ..."] = attributes
                elif attributes["type"] == "dict":
                    if param.annotation is UrlEncodeForm:
                        ps = form_params
                    else:
                        ps = params
                    self._enrich_params_from_example(attributes, ps, name)
                else:
                    params[name] = attributes

            elif issubclass(param.annotation, Type):
                model_params = self._get_params_attr(param, handler.__doc__)
                # 用户指定了从表单中过来的model而不是json
                if model_params["type"] == "form":
                    model_params["type"] = self._get_module_class_name(
                        param.annotation)
                    self._inline_example(model_params)
                    form_params[param.name] = model_params
                else:
                    json_params = model_params
                    json_params["model_type"] = self._get_module_class_name(
                        param.annotation)
                structure = dict()
                type_info.update(
                    self._get_type_info(param.annotation, structure))

        if params:
            info["params"] = params
        if form_params:
            info["form_params"] = form_params
        if path_params:
            info["path_params"] = path_params
        if json_params:
            info["json_params"] = json_params

        if sign.return_annotation is not inspect._empty:
            info["return_type"] = self._get_module_class_name(
                sign.return_annotation)
            # TODO 这个暂时没有什么用处
            structure = dict()
            if issubclass(sign.return_annotation, Type):
                type_info.update(
                    self._get_type_info(sign.return_annotation, structure))

        return_ex = self._extract_return_example(handler.__doc__)

        for ret_ex in return_ex:
            if ret_ex.strip().startswith("`"):
                info.setdefault("return_example", []).append(ret_ex)
            else:
                info.setdefault("return_desc", []).append(ret_ex)
        info["type_info"] = type_info
        # 获取返回响应信息
        return_wrapped = getattr(handler, "__return_wrapped", None)

        if return_wrapped:
            info["resp_info"] = OrderedDict(
                self._extract_resp_info(return_wrapped))

            if "return_type" in info:
                info["return_type"] = '{"code": %s, "%s": %s, "message": "xx"}' \
                                      % (return_wrapped["success_code"],
                                         return_wrapped["success_key_name"],
                                         info["return_type"])
        return info

    @classmethod
    def _extract_param_desc(cls, docstring, param_name):
        regex = cls.regex_format(type="param", param_name=param_name)
        mth = re.search(regex, docstring, re.DOTALL)

        if mth:
            return mth.group(1).strip()

    @classmethod
    def _extract_param_type(cls, docstring, param_name):
        regex = cls.regex_format(type="type", param_name=param_name)
        mth = re.search(regex, docstring, re.DOTALL)

        if mth:
            return mth.group(1).strip()

    def _extract_param_example(self, docstring, param_name):
        regex = self.regex_format("ex", param_name, "(\s*?`+[^`]+?`+)")
        examples = [self._adjust_example(e) for e in re.findall(
            regex, docstring, re.DOTALL)]
        # 为短示例增加序号
        if len(examples) > 1 and "```" not in examples[0]:
            examples = ["%d. %s" % (i + 1, self._adjust_example(e))
                        for i, e in enumerate(examples)]
        return examples

    @classmethod
    def _extract_comment(cls, docstring):
        if not docstring:
            return ""

        mth = re.search(rf"^\s+(.*?){cls.regex_tail}", docstring, re.DOTALL)

        if mth:
            return mth.group(1).strip()

    def _get_params_attr(self, param, doc):
        attributes = dict()

        if param.default != inspect._empty:
            attributes["default"] = param.default
        attributes["type"] = self.param_annotation_mapping.get(
            param.annotation, param.annotation.__name__)

        if doc:
            attributes["type"] = self._extract_param_type(
                doc, param.name) or attributes["type"]
            desc = self._extract_param_desc(doc, param.name)
            if desc:
                attributes["desc"] = desc.strip()
            attributes["examples"] = self._extract_param_example(doc,
                                                                param.name)

        return attributes

    def _extract_return_example(self, docstring):
        if not docstring:
            return []

        return [self._adjust_example(e) for e in re.findall(
            r":return:\s+(.+?)\s*(?=(?:(?::return)|$))", docstring,
            re.DOTALL)]

    @cache_method()
    def _extract_type_annotation(self, cls):
        info = dict()

        for prop, validator in cls.validator.properties.items():
            param = DummyParam(
                prop, getattr(validator, "default", inspect._empty),
                validator.__class__)
            attributes = self._get_params_attr(param, cls.__doc__)
            attributes["allow_null"] = validator.allow_null
            info[prop] = attributes
        return info

    @staticmethod
    def _adjust_example(ex):
        ex = ex.strip()
        lines = ex.split("\n")
        if len(lines) > 1:
            lines[-1] = lines[-1].strip()
            ex = "\n".join(lines)
        return ex

    def _get_module_class_name(self, cls):
        if isinstance(cls, typing.GenericMeta):
            return str(cls)
        elif issubclass(cls, Type):
            return f"{cls.__module__}.{cls.__name__}"
        else:
            return cls.__name__

    def _get_type_info(self, cls, structure):
        if issubclass(cls, Type):
            type_info = self._extract_type_annotation(cls)
            structure[cls] = type_info
            yield self._get_module_class_name(cls), type_info
        elif hasattr(cls, "__args__"):
            child = structure.setdefault(cls, OrderedDict())
            for cls in getattr(cls, "__args__"):
                yield from self._get_type_info(cls, child)
        else:
            type_info = dict()
            structure[cls] = type_info
            yield self._get_module_class_name(cls), type_info

    @staticmethod
    def _extract_resp_info(return_wrapped):
        yield return_wrapped["success_code"], "返回成功"
        for item in (return_wrapped.get("error_info") or {}).items():
            yield item

    def _enrich_params_from_example(self, attributes, params, name):
        examples = attributes.get("examples")

        for example in examples:
            try:
                ex_data = json.loads(example.strip("`json"))

                for ex_name, ex in ex_data.items():
                    exs = params.setdefault(
                        ex_name, dict()).setdefault("examples", [])
                    params.setdefault(ex_name, dict())["type"] = type(
                        ex).__name__
                    exs.append(f"`{json.dumps(ex)}`")

            except json.JSONDecodeError as e:
                print(f"Example format error "
                      f"of {attributes['endpoint']} :{name}")
                raise e

    @staticmethod
    def _inline_example(params):
        for index, ex in enumerate(params.get("examples", [])[:]):
            ex = '`' + ex.strip("`json").replace("\n", " ").strip() + '`'
            params.get("examples")[index] = ex
